/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For 
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 * 
 * $Id$
 */
package org.dataone.tidy.merge.strategy;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Transformer;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.SerializationUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.configuration.Settings;
import org.dataone.service.exceptions.InvalidToken;
import org.dataone.service.exceptions.NotAuthorized;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.NotImplemented;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.v1.AccessPolicy;
import org.dataone.service.types.v1.AccessRule;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.Permission;
import org.dataone.service.types.v1.Replica;
import org.dataone.service.types.v1.ReplicationPolicy;
import org.dataone.service.types.v1.ReplicationStatus;
import org.dataone.service.types.v1.Subject;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.service.util.TypeMarshaller;
import org.dataone.tidy.InclusionsFilter;
import org.dataone.tidy.merge.AuthMNSystemMetadataProxy;
import org.dataone.tidy.merge.MNSystemMetadataDelegate;
import org.dataone.tidy.merge.MergeReport;
import org.dataone.tidy.merge.MergeStrategy;
import org.jibx.runtime.JiBXException;

/**
 * Based on: http://mule1.dataone.org/OperationDocs/20131126_cn_audit_design/cn_systemmeta_audit_activity_diagrams.html
 * Figure 2.1 Merge SystemMetadata
 *
 * @author sroseboo
 * @author waltz
 * @author rnahf
 *
 */
public class MergeSystemMetadata implements MergeStrategy<Map<String, SystemMetadata>> {

    private static Log logger = LogFactory.getLog(MergeSystemMetadata.class);
    //    private static AuditWarningLogger warnLog = AuditWarningLogger.getInstance();
    private MergeReport mergeReport;
    private MNSystemMetadataDelegate mnSysMetaDelegate;
    private static final boolean USE_MN_ACCESSPOLICY = 
    Settings.getConfiguration().getBoolean("tidy.merge.strategy.use_mn_accessPolicy", true);

    
    public MergeSystemMetadata(MergeReport report) {
        
        this.mergeReport = report;
        this.mnSysMetaDelegate = new AuthMNSystemMetadataProxy();
        
    }

    public MergeSystemMetadata(MergeReport report, MNSystemMetadataDelegate mnSysMetaDelegate) {

        this.mergeReport = report;
        this.mnSysMetaDelegate = mnSysMetaDelegate;
    }
    
    /**
     * MergeStrategy implementation for merging systemMetadata on CNs that have become
     * out of sync.  It chooses the 'ascendent' system metadata (latest version and
     * modified date), and merges field values according to the rules in the 
     * protected merge methods of this class.
     * 
     * It records the mergedSystemMetadata in the returned MergeReport, but does
     * so only if there hasn't been a processing exception (something that prevents
     * the merge strategy from completing)
     * 
     * Exceptions are caught and added to the MergeReport, to keep the interface
     * clean.
     */
    @Override
    public MergeReport mend(Map<String, SystemMetadata> sysmetas) {

        try {
            InclusionsFilter inclusionsFilter = InclusionsFilter.getInstance();
            String ascendentSysmeta = determineAscendentSystemMetadata(sysmetas);
            
            if (ascendentSysmeta == null || ascendentSysmeta.isEmpty()) {
                logger.warn("ascendentSysmeta from determineAscendentSystemMetadata is null or empty");
            } else {
                logger.info("determined that " + ascendentSysmeta + " is ascendent from ");
                for (String key : sysmetas.keySet()) {
                    logger.info(key);
                }
            }
            // need to make sure that the passed in systemMetadata is not altered,
            // so substituting a clone of it, once we know what it is :-)
            SystemMetadata copyAscendent = clone(sysmetas.get(ascendentSysmeta));

            if (sysmetas.size() > 1) {

                applyLatestDateSystemMetadataModified(ascendentSysmeta, sysmetas);

                applyArchivedIfAnyArchived(ascendentSysmeta, sysmetas);

                useAscendentsObsoletedBy(ascendentSysmeta, sysmetas);
                useAscendentsObsoletes(ascendentSysmeta, sysmetas);

                mergeAuthoritativeMemberNode(ascendentSysmeta, sysmetas);

                mergeReplicaList(ascendentSysmeta, sysmetas);

                mergeReplicaPolicyInfo(ascendentSysmeta, sysmetas);

                if (isReplicationAllowed(sysmetas.get(ascendentSysmeta))) {
                    collateAllPreferredLists(ascendentSysmeta, sysmetas);
                    mergeBlocked(ascendentSysmeta, sysmetas);
                }
            }
            if (USE_MN_ACCESSPOLICY) {
            	Identifier pid = this.mergeReport.getIdentifier();
                NodeReference authMN = sysmetas.get(ascendentSysmeta).getAuthoritativeMemberNode();
                if (inclusionsFilter.shouldInclude(pid, authMN)) {
      	
            		acceptAuthMNAccessPolicy(ascendentSysmeta, sysmetas);
            	
            	} else {
            		logger.info("AuthMN not called for pid: " + pid.getValue() + " because the identifier is in the bypass list");
            	} 
            } else {
            	// mergeRightsHolder(ascendentSysmeta, sysmetas);
                // mergeAccessPolicy(ascendentSysmeta, sysmetas);
            }

            // this will only get set if otherwise no exceptions
            this.mergeReport.setResultingSystemMetadata(sysmetas.get(ascendentSysmeta));
            // replace the original version of the modified ascendent system metadata in the map for reporting.
            sysmetas.put(ascendentSysmeta, copyAscendent);

        } catch (ServiceFailure e) {
            this.mergeReport.setProcessingException(e);
        } catch (InvalidToken e) {
            this.mergeReport.setProcessingException(e);
        } catch (NotAuthorized e) {
            this.mergeReport.setProcessingException(e);
        } catch (NotImplemented e) {
            this.mergeReport.setProcessingException(e);
        } catch (NotFound e) {
            this.mergeReport.setProcessingException(e);
        } catch (IOException e) {
            this.mergeReport.setProcessingException(e);
        } catch (InstantiationException e) {
            this.mergeReport.setProcessingException(e);
        } catch (IllegalAccessException e) {
            this.mergeReport.setProcessingException(e);
        } catch (JiBXException e) {
            this.mergeReport.setProcessingException(e);
        }

        return this.mergeReport;
    }

    private boolean isReplicationAllowed(SystemMetadata systemMetadata) {
        return systemMetadata.getReplicationPolicy() != null
                && BooleanUtils.isTrue(systemMetadata.getReplicationPolicy()
                        .getReplicationAllowed());

    }

    /**
     * The ascendent sysmeta is the one with the highest serialVersion and then 
     * if more than one at the highest version, the latest modification date.
     * In the case of a tie in both criteria, the first one encountered is used,
     * and a log.warn is thrown stating multiple suitable candidates.
     *  
     * @param sysmetaToMerge
     * @return
     */
    protected String determineAscendentSystemMetadata(Map<String, SystemMetadata> sysmetaToMerge) {
        String candidateAscendent = null;
        int countOfHighestSerial = 0;
        for (Entry<String, SystemMetadata> sysMeta : sysmetaToMerge.entrySet()) {
            if (candidateAscendent == null) {
                candidateAscendent = sysMeta.getKey();
                countOfHighestSerial = 1;
            } else if (sysmetaToMerge.get(candidateAscendent).getSerialVersion()
                    .compareTo(sysMeta.getValue().getSerialVersion()) < 0) {
                candidateAscendent = sysMeta.getKey();
                countOfHighestSerial = 1;
            } else if (sysmetaToMerge.get(candidateAscendent).getSerialVersion()
                    .compareTo(sysMeta.getValue().getSerialVersion()) == 0) {
                if (sysmetaToMerge.get(candidateAscendent).getDateSysMetadataModified()
                        .before(sysMeta.getValue().getDateSysMetadataModified())) {
                    candidateAscendent = sysMeta.getKey();
                    countOfHighestSerial = 1;
                } else if (sysmetaToMerge.get(candidateAscendent).getDateSysMetadataModified()
                        .equals(sysMeta.getValue().getDateSysMetadataModified())) {
                    countOfHighestSerial++;
                } else {
                    ; // same serialVersion, but earlier date, so ignore
                }

            }
            if (candidateAscendent == null || candidateAscendent.isEmpty()) {
                logger.warn("In determineAscendentSystemMetadata, candidateAscendent was assigned from key but is null or empty still");
            }
        }
        //        if (countOfHighestSerial > 1) {
        //            logger.warn(countOfHighestSerial + " systemMetadata objects with same "
        //                    + "serialVersion and modDate were found.  Using the first encountered");
        //        }
        return candidateAscendent;
    }

    /**
     * applies the latest modification date into the mergeTarget
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void applyLatestDateSystemMetadataModified(String mergeTarget,
            Map<String, SystemMetadata> sysmetas) {

        Date modDate = null;
        for (Entry<String, SystemMetadata> sysMeta : sysmetas.entrySet()) {

            if (modDate == null) {
                modDate = sysMeta.getValue().getDateSysMetadataModified();
            } else if (modDate.before(sysMeta.getValue().getDateSysMetadataModified())) {
                modDate = sysMeta.getValue().getDateSysMetadataModified();
            }
        }
        sysmetas.get(mergeTarget).setDateSysMetadataModified(modDate);
    }

    /**
     * sets archived flag to true on the mergeTarget if one of the sysmeta
     * instances have the archived flag set to true.
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void applyArchivedIfAnyArchived(String mergeTarget,
            Map<String, SystemMetadata> sysmetas) {

        if (BooleanUtils.isNotTrue(sysmetas.get(mergeTarget).getArchived())) {
            // then it's either false or null (which are equivalent)

            for (Entry<String, SystemMetadata> sysMeta : sysmetas.entrySet()) {
                if (BooleanUtils.isTrue(sysMeta.getValue().getArchived())) {
                    sysmetas.get(mergeTarget).setArchived(Boolean.TRUE);
                    break;
                }
            }
        } // else, it's already true so don't need to do anything
    }

    /**
     * Take the ascendent sysmeta's obsoletedBy value, unless null, in which
     * case, take the value from one of the other sysmetas.
     * If there are conflicting values among the non-ascendents, log an ERROR
     * and choose the first non-null value encountered in the sysmeta list.
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void useAscendentsObsoletedBy(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        List<Identifier> prioritizationList = new ArrayList<Identifier>();
        // put the mergeTarget value at index 0  (allow null)
        prioritizationList.add(sysmetas.get(mergeTarget).getObsoletedBy());

        // add other unique values to the list if they are not null
        for (Entry<String, SystemMetadata> sysmeta : sysmetas.entrySet()) {
            Identifier obsoletedBy = sysmeta.getValue().getObsoletedBy();
            if (obsoletedBy != null && !prioritizationList.contains(obsoletedBy)) {
                prioritizationList.add(sysmeta.getValue().getObsoletedBy());
            }
        }

        // set the value in the mergeTarget
        if (prioritizationList.get(0) == null && prioritizationList.size() > 1) {
            sysmetas.get(mergeTarget).setObsoletedBy(prioritizationList.get(1));

        } // else accept the existing value in the mergeTarget

        // now check for and report (log) conflicts

        if (prioritizationList.get(0) == null) {
            prioritizationList.remove(0);
        }
        if (prioritizationList.size() > 1) {
            // there's a conflict that we need to resolve

            StringBuffer sb = new StringBuffer();
            for (Identifier id : prioritizationList) {
                sb.append(id.getValue());
                sb.append(" ");
            }
            mergeReport.addWarning("Multiple values found for ObsoletedBy field: " + sb.toString()
                    + ": Using the first one.");
        }
    }

    /**
     * Take the ascendent sysmeta's obsoletes value, unless null, in which
     * case, take the value from one of the other sysmetas.
     * If there are conflicting values among the non-ascendents, log an ERROR
     * and choose the first non-null value encountered in the sysmeta list.
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void useAscendentsObsoletes(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        List<Identifier> prioritizationList = new ArrayList<Identifier>();
        // put the mergeTarget value at index 0
        prioritizationList.add(sysmetas.get(mergeTarget).getObsoletes());
        // add other values to the list if they are not null
        for (Entry<String, SystemMetadata> sysmeta : sysmetas.entrySet()) {
            Identifier obsoletes = sysmeta.getValue().getObsoletes();
            if (obsoletes != null && !prioritizationList.contains(obsoletes)) {
                prioritizationList.add(sysmeta.getValue().getObsoletes());
            }
        }

        // set the value in the mergeTarget
        if (prioritizationList.get(0) == null && prioritizationList.size() > 1) {
            sysmetas.get(mergeTarget).setObsoletes(prioritizationList.get(1));
        } else {
            // accept the existing value in the mergeTarget
        }

        // now check for and report (log) conflicts

        if (prioritizationList.get(0) == null) {
            prioritizationList.remove(0);
        }
        if (prioritizationList.size() > 1) {
            // there's a conflict that we need to resolve

            StringBuffer sb = new StringBuffer();
            for (Identifier id : prioritizationList) {
                sb.append(id.getValue());
                sb.append(" ");
            }
            mergeReport.addWarning("Multiple values found for Obsoletes field: " + sb.toString()
                    + ": Using the first one.");
        }
    }

    /**
     * Make a list of all replicas encountered.  For multiple replica records for
     * the same membernode, choose the one with the latest verification date and
     * status.  (status is the tie-breaker)
     * 
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void mergeReplicaList(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        List<Replica> replicas = gatherUniqueReplicas(sysmetas);
        sysmetas.get(mergeTarget).clearReplicaList();
        sysmetas.get(mergeTarget).setReplicaList(replicas);
    }

    /*
     * Creates a super-list of replicas from the individual sysmetas.  When a replica
     * is found in more than one sysmeta, the one with latest verification date and then highest
     * priority is used.
     */
    private List<Replica> gatherUniqueReplicas(Map<String, SystemMetadata> systemMetadataList) {
        Map<String, Replica> replicasByNodeId = new HashMap<String, Replica>();
        for (SystemMetadata sysMeta : systemMetadataList.values()) {
            if (CollectionUtils.isNotEmpty(sysMeta.getReplicaList())) {
                for (Replica replica : sysMeta.getReplicaList()) {
                    String nodeId = replica.getReplicaMemberNode().getValue();
                    if (replicasByNodeId.containsKey(nodeId) == false) {
                        replicasByNodeId.put(nodeId, replica);
                    } else {
                        Replica currentMerged = replicasByNodeId.get(nodeId);
                        ReplicationStatus currentMergedStatus = currentMerged
                                .getReplicationStatus();

                        // 2 replica for the same target node
                        // keep the replica with newest replica verified date
                        //  if the date is the same keep highest priority status.
                        if (replica.getReplicaVerified().after(currentMerged.getReplicaVerified())) {
                            replicasByNodeId.put(nodeId, replica);

                        } else if (replica.getReplicaVerified().equals(
                                currentMerged.getReplicaVerified())) {

                            int currentPriority = MergeSystemMetadata.replicaStatusPriorityMap
                                    .get(currentMergedStatus);
                            int replicaPriority = MergeSystemMetadata.replicaStatusPriorityMap
                                    .get(replica.getReplicationStatus());
                            if (replicaPriority > currentPriority) {
                                replicasByNodeId.put(nodeId, replica);
                            }
                        }

                    }
                }
            }
        }
        List<Replica> replicas = new ArrayList<Replica>();
        replicas.addAll(replicasByNodeId.values());
        return replicas;
    }

    /**
     * merges the isReplicationAllowed and NumberReplicas values.
     * isReplicationAllowed value  = OR(isReplicationAllowed of sysmetas)
     * numberReplicas = max(numberReplicas of sysmetas) 
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    // TODO: replace with "use_CN_ORC_ReplicaPolicyInfo"
    protected void mergeReplicaPolicyInfo(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        boolean replicationAllowed = isReplicationAllowed(sysmetas.get(mergeTarget));
        if (replicationAllowed == false) {
            for (SystemMetadata sysMeta : sysmetas.values()) {
                if (isReplicationAllowed(sysMeta)) {
                    if (sysmetas.get(mergeTarget).getReplicationPolicy() == null) {
                        sysmetas.get(mergeTarget).setReplicationPolicy(new ReplicationPolicy());
                    }
                    sysmetas.get(mergeTarget).getReplicationPolicy()
                            .setReplicationAllowed(Boolean.TRUE);
                    replicationAllowed = true;
                    break;
                }
            }
        }

        if (replicationAllowed) {
            Integer numberReplicas = Integer.valueOf(0);
            for (SystemMetadata sysMeta : sysmetas.values()) {
                if (sysMeta.getReplicationPolicy() != null
                        && sysMeta.getReplicationPolicy().getNumberReplicas() != null
                        && sysMeta.getReplicationPolicy().getNumberReplicas()
                                .compareTo(numberReplicas) > 0) {
                    numberReplicas = sysMeta.getReplicationPolicy().getNumberReplicas();
                }
            }
            sysmetas.get(mergeTarget).getReplicationPolicy().setNumberReplicas(numberReplicas);
        }
    }

    /**
     * The Preferred lists are ordered, and so when merging, the resulting superset
     * needs to maintain the original ordering as best it can, and this is done 
     * via collating the lists, as in: 
     *    [a1,a2,a3...], [b1,b2...], [c1,c2,..] => [a1,b1,c1,a2,b2,c2,...]
     *
     * The list of the ascendent (target) sysmeta is the first list in the order 
     * of starting lists, so in the above example a is assumed to be from the 
     * ascendent list.
     * 
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void collateAllPreferredLists(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        LinkedList<NodeReference> mergedPreferredNodeList = new LinkedList<NodeReference>();
        Boolean modifiedPreferredList = false;
        // The order of the lists must be collated 
        // The nodes will have to be interleaved
        // so that any non duplicate similarly ranked items should follow each other
        // with the mergeTarget taking priority.

        Map<Integer, LinkedList<String>> orderedMapofNodeLists = new HashMap<Integer, LinkedList<String>>();

        LinkedList<String> targetPreferredStringList = new LinkedList<String>();

        CollectionUtils.collect(sysmetas.get(mergeTarget).getReplicationPolicy()
                .getPreferredMemberNodeList(), NodeReferenceToStringTransformer.getInstance(),
                targetPreferredStringList);
        for (int i = 0; i < targetPreferredStringList.size(); ++i) {
            LinkedList<String> nodeList = new LinkedList<String>();
            nodeList.offer(targetPreferredStringList.get(i));
            orderedMapofNodeLists.put(i, nodeList);
        }
        // this should become a Map of CN ids to addedNodesToReport list for accurate reporting
        List<String> addedNodesToReport = new ArrayList<String>();
        for (SystemMetadata metadataProvided : sysmetas.values()) {
            LinkedList<String> providedPreferredStringList = new LinkedList<String>();

            CollectionUtils.collect(metadataProvided.getReplicationPolicy()
                    .getPreferredMemberNodeList(), NodeReferenceToStringTransformer.getInstance(),
                    providedPreferredStringList);
            for (int i = 0; i < providedPreferredStringList.size(); ++i) {
                String providedNode = providedPreferredStringList.get(i);
                if (!targetPreferredStringList.contains(providedNode)) {
                    modifiedPreferredList = true;
                    addedNodesToReport.add(providedNode);
                    targetPreferredStringList.add(providedNode);
                    if (orderedMapofNodeLists.containsKey(i)) {
                        orderedMapofNodeLists.get(i).offer(providedNode);
                    } else {
                        LinkedList<String> nodeList = new LinkedList<String>();
                        nodeList.offer(providedPreferredStringList.get(i));
                        orderedMapofNodeLists.put(i, nodeList);
                    }

                }
            }
        }
        if (modifiedPreferredList) {
            String addedNodes = StringUtils.join(addedNodesToReport, ',');
            //            this.mergeReport.addWarning("Preferred will add  " + addedNodes
            //                    + " found in the PreferredNodeList");
            List<Integer> orderedListOfEntries = new ArrayList<Integer>();
            orderedListOfEntries.addAll(orderedMapofNodeLists.keySet());
            Collections.sort(orderedListOfEntries);
            for (Integer i : orderedListOfEntries) {
                List<String> preferredNodeList = orderedMapofNodeLists.get(i);
                for (String node : preferredNodeList) {
                    NodeReference nodeReference = new NodeReference();
                    nodeReference.setValue(node);
                    mergedPreferredNodeList.offer(nodeReference);
                }
            }
            sysmetas.get(mergeTarget).getReplicationPolicy()
                    .setPreferredMemberNodeList(mergedPreferredNodeList);
        }
    }

    /**
     * Create a superset of blocked member nodes from the different systemMetadatas,
     * logging differences between the mergeTarget and the others (additions from 
     * the others)
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void mergeBlocked(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        List<String> targetBlockedStringList = new ArrayList<String>();
        Boolean modifiedBlockedList = false;
        LinkedList<NodeReference> mergedBlockedNodeList = new LinkedList<NodeReference>();

        CollectionUtils.collect(sysmetas.get(mergeTarget).getReplicationPolicy()
                .getBlockedMemberNodeList(), NodeReferenceToStringTransformer.getInstance(),
                targetBlockedStringList);
        for (SystemMetadata sysmeta : sysmetas.values()) {
            List<String> providedBlockedStringList = new ArrayList<String>();

            CollectionUtils.collect(sysmeta.getReplicationPolicy().getBlockedMemberNodeList(),
                    NodeReferenceToStringTransformer.getInstance(), providedBlockedStringList);
            providedBlockedStringList.removeAll(targetBlockedStringList);
            if (!providedBlockedStringList.isEmpty()) {
                modifiedBlockedList = true;
                // the targetBlockedNodeList does not contain values that are
                // a part of the providedBlockedNodeList, so report the additions
                String addedNodes = StringUtils.join(providedBlockedStringList, ',');
                //                this.mergeReport.addWarning("BlockedNodeList will add  " + addedNodes
                //                        + " found in the BlockedNodeList");
                Collection<String> newNodeList = CollectionUtils.union(targetBlockedStringList,
                        providedBlockedStringList);
                logger.debug("newNodeList size: " + newNodeList.size());
                targetBlockedStringList.clear();
                targetBlockedStringList.addAll(newNodeList);
            }
        }
        if (modifiedBlockedList) {
            logger.debug("blockList was modified");
            for (String node : targetBlockedStringList) {
                NodeReference nodeReference = new NodeReference();
                nodeReference.setValue(node);
                mergedBlockedNodeList.add(nodeReference);
                logger.debug("New Block list contains: " + node);
            }

            sysmetas.get(mergeTarget).getReplicationPolicy()
                    .setBlockedMemberNodeList(mergedBlockedNodeList);
        }
    }

    /**
     * Calls out to the AuthoritativeMN to get it's accessPolicy.  If it cannot 
     * be retrieved due to a NotFound, simply use what's already in the ascendent
     * systemMetadata.
     * @param mergeTarget
     * @param sysmetas 
     * @throws ServiceFailure 
     * @throws InvalidToken 
     * @throws NotAuthorized 
     * @throws NotImplemented 
     */
    protected void acceptAuthMNAccessPolicy(String mergeTarget, Map<String, SystemMetadata> sysmetas)
            throws ServiceFailure, InvalidToken, NotAuthorized, NotImplemented {

        NodeReference authMN = sysmetas.get(mergeTarget).getAuthoritativeMemberNode();
        Identifier pid = sysmetas.get(mergeTarget).getIdentifier();
        this.mergeReport.getMergeResult().setAuthoritativeMemberNode(authMN.getValue());
        try {
            SystemMetadata authMNSysmeta = this.mnSysMetaDelegate.getSystemMetadata(authMN, pid);
            AccessPolicy ap = (AccessPolicy) SerializationUtils.clone(authMNSysmeta
                    .getAccessPolicy());
            sysmetas.get(mergeTarget).setAccessPolicy(ap);
            this.mergeReport.setAuthMnSystemMetadata(authMNSysmeta);
        } catch (InvalidToken e) {
            this.mergeReport.addWarning("Exception retrieving AuthMN systemMetadata: "
                    + e.getClass().getSimpleName() + ": " + e.getDescription());
            throw e;
        } catch (NotAuthorized e) {
            this.mergeReport.addWarning("Exception retrieving AuthMN systemMetadata: "
                    + e.getClass().getSimpleName() + ": " + e.getDescription());
            throw e;
        } catch (NotImplemented e) {
            this.mergeReport.addWarning("Exception retrieving AuthMN systemMetadata: "
                    + e.getClass().getSimpleName() + ": " + e.getDescription());
            throw e;
        } catch (ServiceFailure e) {
            this.mergeReport.addWarning("Exception retrieving AuthMN systemMetadata: "
                    + e.getClass().getSimpleName() + ": " + e.getDescription());
            throw e;
        } catch (NotFound e) {
            this.mergeReport.addWarning("Exception retrieving AuthMN systemMetadata: "
                    + e.getClass().getSimpleName() + ": " + e.getDescription());
        }

    }

    /**
     * The Access Policy of the ascendent (target) sysmeta is used in all cases,
     * and discrepancies, between it and other versions are logged
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void mergeAccessPolicy(String mergeTarget, Map<String, SystemMetadata> sysmetas) {
        // the mergeTarget Wins everytime!
        // But we need to log, or provide notification if there are discrepancies
        // from the other two systemMetadata

        // the access Policy of the various documents should be ordered.
        // Any variations of the access policies from the policy of the
        // ascendent version must be logged
        Map<Subject, List<String>> targetSubjectPermissions = buildSubjectPermissionsHashMap(sysmetas
                .get(mergeTarget).getAccessPolicy().getAllowList());

        for (SystemMetadata metadataProvided : sysmetas.values()) {
            Map<Subject, List<String>> providedSubjectPermissions = buildSubjectPermissionsHashMap(metadataProvided
                    .getAccessPolicy().getAllowList());
            for (Subject targetSubject : targetSubjectPermissions.keySet()) {
                if (providedSubjectPermissions.containsKey(targetSubject)) {
                    // find out if there is any difference between the target and the provided lists
                    Collection<String> solution = CollectionUtils.disjunction(
                            targetSubjectPermissions.get(targetSubject),
                            providedSubjectPermissions.get(targetSubject));
                    if (!solution.isEmpty()) {
                        // report the permissions that will be inserted or deleted by the target 
                        // SystemMetadata as viewed from the CN's provided permissions
                        solution = CollectionUtils.subtract(
                                targetSubjectPermissions.get(targetSubject),
                                providedSubjectPermissions.get(targetSubject));
                        if (!solution.isEmpty()) {
                            // the targetSubjectPermissions contains values that are not
                            // a part of the providedSubjectPermissions, so report the addition
                            String addPermissions = StringUtils.join(solution, ',');
                            //                            this.mergeReport.addWarning(targetSubject.getValue() + 
                            //                           		" will have permissions  " + 
                            //                                   addPermissions + " added to the AccessPolicy");
                        }

                        solution = CollectionUtils.subtract(
                                providedSubjectPermissions.get(targetSubject),
                                targetSubjectPermissions.get(targetSubject));
                        if (!solution.isEmpty()) {
                            // the targetSubjectPermissions does not contain values that are
                            // a part of the providedSubjectPermissions, so report the subtraction
                            String minusPermissions = StringUtils.join(solution, ',');
                            //                            this.mergeReport.addWarning(targetSubject.getValue() + 
                            //                            		" has removed permissions  " +
                            //                                    minusPermissions + " found in the AccessPolicy");
                        }
                    }
                    //                } else {
                    //                    String missingPermissions = StringUtils.join(
                    //                            providedSubjectPermissions.get(targetSubject), ',');
                    //                    this.mergeReport.addWarning(targetSubject.getValue() + " with permissions " + 
                    //                            missingPermissions  + " was missing from the AccessPolicy");
                }
            }
        }
    }

    /**
     * The RightsHolder of the ascendent (target) sysmeta is used in all cases,
     * and discrepancies, between it and other versions are logged.
     * 
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    //    protected SystemMetadata mergeRightsHolder(String mergeTarget,
    //            Map<String, SystemMetadata> sysmetas) {
    //        // the mergeTarget Wins everytime!
    //        // But we need to log, or provide notification if there are discrepancies
    //        // from the other two systemMetadata
    //        mergeTarget.getRightsHolder();
    //        for (SystemMetadata origin : sysmetas) {
    //            if (!mergeTarget.getRightsHolder().equals(origin.getRightsHolder())) {
    //                warnLog.warn(mergeTarget, "Rightsholder value:"
    //                        + origin.getRightsHolder().getValue() + " replaced with:"
    //                        + mergeTarget.getRightsHolder().getValue());
    //            }
    //        }
    //        return mergeTarget;
    //    }

    /**
     * The AuthoritativeMember Node of the ascendent (target) sysmeta is used in all cases,
     * and discrepancies, between it and other versions are logged.
     *
     * @param mergeTarget
     * @param sysmetas
     * @return
     */
    protected void mergeAuthoritativeMemberNode(String mergeTarget,
            Map<String, SystemMetadata> sysmetas) {
        // the mergeTarget Wins everytime!
        // But we need to log, or provide notification if there are discrepancies
        // from the other two systemMetadata
        sysmetas.get(mergeTarget).getAuthoritativeMemberNode();
        for (SystemMetadata origin : sysmetas.values()) {
            if (!sysmetas.get(mergeTarget).getAuthoritativeMemberNode()
                    .equals(origin.getAuthoritativeMemberNode())) {
                this.mergeReport.addWarning("AuthoritativeMemberNode value:"
                        + origin.getAuthoritativeMemberNode().getValue() + " replaced with:"
                        + sysmetas.get(mergeTarget).getAuthoritativeMemberNode().getValue());
            }
        }
    }

    private Map<Subject, List<String>> buildSubjectPermissionsHashMap(List<AccessRule> allowList) {
        Map<Subject, List<String>> subjectPermissionsMap = new HashMap<Subject, List<String>>();

        for (AccessRule accessRule : allowList) {
            // convert all the subjects in to a hash map with all permissions
            // transformed to a list of Strings
            List<Permission> permissionList = accessRule.getPermissionList();
            List<Subject> subjectList = accessRule.getSubjectList();
            // for each subject, create a list of permissions. 
            for (Subject subject : subjectList) {
                List<String> permissionStringsList = new LinkedList<String>();
                CollectionUtils.collect(permissionList,
                        PermissionToStringTransformer.getInstance(), permissionStringsList);

                if (subjectPermissionsMap.containsKey(subject)) {
                    subjectPermissionsMap.get(subject).addAll(permissionStringsList);
                    Collections.sort(subjectPermissionsMap.get(subject));
                } else {
                    Collections.sort(permissionStringsList);
                    subjectPermissionsMap.put(subject, permissionStringsList);
                }
            }

        }
        return subjectPermissionsMap;
    }

    /**
     * a custom implementation of deep-cloning of sysmeta through serialization
     * It should be part of TypeMarshaller in d1_common_java, but someday...
     * @param sysmeta
     * @return
     * @throws IOException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws JiBXException
     */
    private SystemMetadata clone(SystemMetadata sysmeta) throws IOException,
            InstantiationException, IllegalAccessException, JiBXException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            if (baos == null) {
                logger.warn("ByteArrayOutputStream provided a null pointer");
            }
            if (sysmeta == null) {
                logger.warn("sysmeta provided a null pointer");
            }
            TypeMarshaller.marshalTypeToOutputStream(sysmeta, baos);
        } finally {
            if (baos != null)
                try {
                    baos.close();
                } catch (IOException ex) {
                    logger.error(ex.getMessage(), ex);
                }
        }

        InputStream bais = new ByteArrayInputStream(baos.toByteArray());
        try {
            return TypeMarshaller.unmarshalTypeFromStream(SystemMetadata.class, bais);
        } finally {
            try {
                if (bais != null) {
                    bais.close();
                }
            } catch (IOException ex) {
                logger.error(ex.getMessage(), ex);
            }
        }

    }

    private static class SubjectToStringTransformer implements Transformer {

        private static SubjectToStringTransformer subjectToStringTransformer = null;

        public static SubjectToStringTransformer getInstance() {
            if (subjectToStringTransformer == null) {
                subjectToStringTransformer = new SubjectToStringTransformer();
            }
            return subjectToStringTransformer;
        }

        private SubjectToStringTransformer() {

        }

        public Object transform(Object input) {
            Subject subject = (Subject) input;
            return subject.getValue();
        }
    }

    private static class PermissionToStringTransformer implements Transformer {

        private static PermissionToStringTransformer permissionToStringTransformer = null;

        public static PermissionToStringTransformer getInstance() {
            if (permissionToStringTransformer == null) {
                permissionToStringTransformer = new PermissionToStringTransformer();
            }
            return permissionToStringTransformer;
        }

        private PermissionToStringTransformer() {
        }

        public Object transform(Object input) {
            Permission permission = (Permission) input;
            return permission.xmlValue();
        }
    }

    private static class NodeReferenceToStringTransformer implements Transformer {

        private static NodeReferenceToStringTransformer nodeIdentifierToStringTransformer = null;

        public static NodeReferenceToStringTransformer getInstance() {
            if (nodeIdentifierToStringTransformer == null) {
                nodeIdentifierToStringTransformer = new NodeReferenceToStringTransformer();
            }
            return nodeIdentifierToStringTransformer;
        }

        private NodeReferenceToStringTransformer() {
        }

        public Object transform(Object input) {
            NodeReference nodeIdentifier = (NodeReference) input;
            return nodeIdentifier.getValue();
        }
    }

    private static Map<ReplicationStatus, Integer> replicaStatusPriorityMap = new HashMap<ReplicationStatus, Integer>();

    static {
        int valueCounter = 0;
        replicaStatusPriorityMap.put(ReplicationStatus.QUEUED, valueCounter++);
        replicaStatusPriorityMap.put(ReplicationStatus.REQUESTED, valueCounter++);
        replicaStatusPriorityMap.put(ReplicationStatus.FAILED, valueCounter++);
        replicaStatusPriorityMap.put(ReplicationStatus.COMPLETED, valueCounter++);
        replicaStatusPriorityMap.put(ReplicationStatus.INVALIDATED, valueCounter++);
    }
}
